package cede;


/**
 * Handles the computation of given number of centroid factors and loading vectors
 *
 */

public class CentroidDecomposition {

	public Matrix b;
	public Matrix v;
	public static boolean shouldLog = false;

	//Comp complexity: l*((mn+m)+(mn+n)+(m+2mn+x*(mn + 6m)) + mn + 2m + mn + mn) = l*(x*mn + x*6m + 7mn + 4m + n) steps
	//in worst case: x=m ==> Comp complexity: l*(mmn + 6mm + 7mn + 4m + n) steps


	// Algorithm of the decomposition
	public void decompose(Matrix a, int numberOfFactors) { 

		log("A" + a);
		Matrix bi = new Matrix(1,1);
		Matrix vi = new Matrix(1,1);


		int iteration = 0;
		log("Starting decomposition...");

		do {
			log("Iteration " + iteration);		

			if (iteration==1) {
				this.b = bi.clone();
				this.v = vi.clone();
			}
			else if (iteration>1) {
				this.b = this.b.appendRight(bi);  //= m*(n+1) steps = O(mn+m)
				this.v = this.v.appendRight(vi);  //= n*(m+1) steps = O(mn+n)
			}

			if(iteration < numberOfFactors) {

				TimeWatch watch = TimeWatch.start();
				log("Computing sign vector z...");			
				Matrix z = findSignVector(a);	// worst case: m + 2mn + x*(mn + 6m)
				log("Find sign vector time: " + watch.time());

				log("Computing c...");			
				Matrix c = a.transposeTimesZ(z); // nm steps = O(nm)

				log("Computing v...");			
				vi = c.times(1/c.norm()); //2m step steps


				log("Computing bi...");			
				bi = a.timesVi(vi);   // nm steps	
			}

			iteration++;

			log("Computing ai+1 (dimension reduction)...");		
			a = a.MinusBitimesViTranspose(bi, vi); //m*n step = O(nm)

		} while (iteration<=numberOfFactors);

		//		System.out.println("\nv:\n" + this.v );
		//		System.out.println("\nb:\n" + this.b );

	}

	//Algorithm of finding the right sign vector
	private  Matrix findSignVector(Matrix a) { //  m + 2mn + x*(mn + 6m), worst case:m + 2mn + m*(mn + 6m)
		TimeWatch watch = TimeWatch.start();
		Matrix z = new Matrix(a.getRows(), 1, 1.0f); //m steps
		log("Z initializing is done in : " + watch.time() + " ms.");

		int i = 0;
		int bestIndex = 0;
		double maxValue = 0.0;

		watch.reset();
		Matrix wi = a.computeInitialWi(); //2mn steps = O(2mn)
		log("wi: " + wi);
		log("Calculating the initial wi is done in : " + watch.time() + " ms ");

		watch.reset();
		do { //worst case : x*(mn + 6m)

			if(i > 0) {
				wi.minuscolP((Matrix) a, bestIndex);  //mn steps = O(mn)
			}

			bestIndex = 0; maxValue = 0.0;


			for (int r=0; r<wi.getRows(); r++) { // 6*m steps = O(m)
				double elementZ = z.get(r, 0);
				double elementWi = wi.get(r, 0);
				if (elementZ*elementWi < 0) {
					if (maxValue < Math.abs(elementWi)) {
						maxValue = Math.abs(elementWi);
						bestIndex = r;
					}
				}
			}

			log("maxvalue:" + maxValue);

			if (maxValue != 0) {
				z.set(bestIndex, 0, -1*z.get(bestIndex,0));
			}
			i++;
		} while (maxValue != 0.0);

		log("Do - While is done in: " + watch.time() + " ms.");
		return z;

	}

	//returns the first column of the matrix, obtained by the dimension reduction
	public Matrix getApprox() {

		Matrix approx = b.times(v.getRow(0).transpose());
		return approx;

	}

	private static void log(String text) {
		if (shouldLog) {
			System.out.println(text);
		}
	}

	public Matrix getB() {
		return b;
	}

	public Matrix getV() {
		return v;
	}

	public static boolean isShouldLog() {
		return shouldLog;
	}


}
