package cede;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.Formatter;

/**
 * This class represents a matrix, and contains all the operations, which are performed by CD
 *
 */

public class Matrix implements Serializable {

	public float[][] matrix;
	int rows, cols;
	private boolean transposed = false;


	private static final long serialVersionUID = -4908840437987432589L;

	//creates a matrix and filled with the initValue
	public Matrix(int rows, int cols, float initValue) {
		this(rows, cols);

		for (int r=0; r<rows; r++) {
			for (int c=0; c<cols; c++) {
				set(r,c,initValue);
			}
		}
	}

	//creates a matrix from a given matrix
	public Matrix(Matrix copyFrom) {
		this(copyFrom.rows, copyFrom.cols);
		this.transposed = copyFrom.transposed;
		for (int r=0; r<rows; r++) {
			for (int c=0; c<cols; c++) {
				set(r,c,copyFrom.get(r,c));
			}
		}
	}

	public Matrix(int rows, int cols) { //m*n steps
		this.rows = rows;
		this.cols = cols;
		this.matrix = new float[rows][cols];
	}

	public void set(int r, int c, float value) {
		this.matrix[r][c]=value;
	}

	public float get(int r, int c) {
		return matrix[r][c];
	}


	public int getRows() {
		return rows;
	}

	public int getCols() {
		return cols;
	}

	public Matrix clone() {
		return new Matrix(this);
	}
	
	public boolean sgn(int row, int col) {
		if(this.get(row,col) < 0)
			return false;
		else
			return true;
	}

	public Matrix appendRight(Matrix a) {  //m*(n+a.n) step 
		if (this.rows!=a.rows) {
			throw new RuntimeException("Dimension mismatch when trying to attach matrices.");
		}

		Matrix attached= new Matrix(rows, this.cols+a.cols);

		for (int r=0; r<rows; r++) {   //m times
			for (int c=0; c<this.cols; c++) {  //
				attached.set(r,c,get(r,c));
			}
			for (int c=0; c<a.cols; c++) {
				attached.set(r,this.cols+c,a.get(r,c));
			}
		}

		return attached;
	}

	public Matrix appendAfter(Matrix a) {
		if (this.cols!=a.cols) {
			throw new RuntimeException("Dimension mismatch when trying to attach matrices.");
		}
		Matrix attached= new Matrix(this.rows + a.rows, this.cols);

		for (int c=0; c<this.cols; c++) {  
			for (int r=0; r<this.rows; r++) {  
				attached.matrix[r][c] = this.matrix[r][c];
			}
			for (int r=0; r<a.rows; r++) {
				attached.matrix[this.rows+r][c] = a.matrix[r][c];
			}
		}

		return attached;
	}

	public Matrix appendReversed(Matrix a) {
		if (this.cols!=a.cols) {
			throw new RuntimeException("Dimension mismatch when trying to attach matrices.");
		}
		Matrix attached= new Matrix(this.rows + a.rows, this.cols);

		for (int c=0; c<this.cols; c++) {  
			for (int r=0; r<this.rows; r++) {  
				attached.matrix[r][c] = this.matrix[r][c];
			}
			for (int r=a.rows-1; r>0; r--) {
				attached.matrix[this.rows+a.rows-r][c] = a.matrix[r][c];
			}
		}

		return attached;
	}

	public void save(String filename) throws FileNotFoundException, IOException {
		ObjectOutputStream objstream;
		objstream = new ObjectOutputStream(new FileOutputStream(filename));
		objstream.writeObject(this);
		objstream.close();
	}

	public static Matrix load(String filename) throws IOException, ClassNotFoundException {

		ObjectInputStream objstream = new ObjectInputStream(new FileInputStream(filename));
		Object object = objstream.readObject();
		objstream.close();
		return (Matrix)object;
	}

	public Matrix times(float a) { //m*n steps O(mn)
		Matrix prod = this.clone();
		for (int r=0; r<this.rows; r++) {
			for (int c=0; c<this.cols; c++) {
				prod.set(r,c,get(r,c) * a);
			}
		}
		return prod;
	}

	public Matrix computeInitialWiByStandardizedData() {
		Matrix wi = new Matrix(this.getRows(), 1);

		float[] sumOfRows = new float[this.getCols()];

		for(int c=0; c<this.getCols(); c++) {
			float temp = 0;
			for(int r=0; r<this.getRows(); r++) {
				temp += this.matrix[r][c];
			}
			sumOfRows[c] = temp;
		}

		for(int r=0; r<wi.getRows(); r++) {
			float[] rowi = this.matrix[r];
			float temp = 0;
			for(int c=0; c<this.getCols(); c++) {
				temp+=rowi[c]*sumOfRows[c];
			}
			wi.matrix[r][0]=temp-1;
		}
		return wi;
	}


	public Matrix computeInitialWi() {   // O(2mn)
		Matrix wi = new Matrix(this.getRows(), 1);

	float[] sumOfRows = new float[this.getCols()];

	for(int c=0; c<this.getCols(); c++) { //n times
		float temp = 0;
		for(int r=0; r<this.getRows(); r++) { //m times
			temp += this.matrix[r][c];
		}
		sumOfRows[c] = temp;
	}

	for(int r=0; r<wi.getRows(); r++) { // m times
		float[] rowi = this.matrix[r];
		float temp = 0;
		float substract = 0;
		for(int c=0; c<this.getCols(); c++) { // n times
			temp+=rowi[c]*sumOfRows[c];
			substract+=rowi[c]*rowi[c];
		}
		temp-=substract;
		wi.matrix[r][0]=temp;
	}
	return wi;
	}

	public void minuscolP(Matrix a, int k) { //mn steps = O(mn)
		float[] rowk = a.matrix[k];

		for(int r = 0; r < this.rows; r++) { // m times
			float substract = 0;
			for(int c = 0; c < a.cols; c++) {  // n times
				substract += a.matrix[r][c]*rowk[c];
			}
			this.matrix[r][0] -= 2*substract;
		}
	} 

	public Matrix transposeTimesZ(Matrix Z) {  // n*m steps = O(n*m)
		Matrix C = new Matrix(this.cols, 1);

		for (int c=0; c<this.cols; c++) {
			float s = 0;
			for(int r=0; r<this.rows; r++) {
				s += this.get(r,c)*Z.get(r,0);
			}
			C.set(c,0,s);    			
		}
		return C;
	}

	public float norm() { // m step
		float norm = 0;
		for (int i=0; i<this.rows; i++)
			norm += this.get(i, 0)*this.get(i, 0);
		norm = (float) Math.sqrt(norm);
		return norm;
	}

	public Matrix timesVi(Matrix Vi) {		//n*m step = O(nm)
		Matrix Bi = new Matrix(this.rows, 1);

		for(int r=0; r<this.rows; r++) {
			float s = 0;
			for(int c=0; c<this.cols; c++) {
				s += this.get(r,c)*Vi.get(c,0);
			}
			Bi.set(r,0,s);
		}
		return Bi;
	}

	public Matrix MinusBitimesViTranspose(Matrix Bi, Matrix Vi) {		//m*n step = O(nm)

		for(int r=0; r<Bi.rows; r++) {
			float s = 0;
			for(int c=0; c<Vi.rows; c++) {
				s = Bi.get(r,0)*Vi.get(c,0);
				this.set(r,c,get(r,c) - s);
			}
		}
		return this;
	}

	public String toString() {
		Formatter f = new Formatter();

		for (int r=0; r<rows; r++) {
			f.format("[");
			for (int c=0; c<cols; c++) {
				f.format("% 12.3f ", get(r,c));
				if (c==cols-1) f.format("]%n");
			}

		}
		return f.toString();
	}

	public void writeInTxt(String filename) throws IOException {
		FileWriter fw = new FileWriter(filename);  
		PrintWriter pw = new PrintWriter(fw); 
		for (int r=0; r<rows; r++) {	
			for (int c=0; c<cols; c++) { 
				pw.print(get(r,c) + "	");
			}
			pw.print("\n");
		}  
		pw.close(); 
	}

	public Matrix getRow(int row) {
		return this.subMatrix(row, row+1, 0, this.cols);
	}

	public Matrix getCol(int col) {
		return this.subMatrix(0, this.rows, col, col+1);
	}

	/**
	 * Extracts a sub-matrix defined by index intervals for rows and columns.
	 * 
	 * @param startRow: first row index to include
	 * @param endRow: first row index NOT to include after startRow
	 * @param startCol: first column index to include
	 * @param endCol: first column index NOT to include after startCol
	 * @return
	 */
	public Matrix subMatrix(int startRow, int endRow, int startCol, int endCol) {
		checkDimensions(startRow, startCol);
		checkDimensions(endRow-1, endCol-1);
		if ((startRow>=endRow) || (startCol>=endCol)) {
			throw new RuntimeException("Invalid sub-matrix indexing: trying to extract zero rows or columns.");

		}

		Matrix subMatrix = new Matrix(endRow-startRow, endCol-startCol);

		for (int r=startRow; r<endRow; r++) {
			for (int c=startCol; c<endCol; c++) {
				subMatrix.matrix[r-startRow][c-startCol]=matrix[r][c];
			}
		}

		return subMatrix;

	}

	private void checkDimensions(int r, int c) {
		if (r>=this.rows || r<0 || c<0 || c>=this.cols) {
			throw new RuntimeException("Invalid indexing: r=" + r + ", c="+c+" in a matrix with "+rows+" rows and "+cols+" columns.");
		}
	}

	public Matrix createHighNMatrix() {
		int m = this.rows/5000;
		Matrix dataMatrix = new Matrix(1,1);

		for(int i = 0; i < m; i++) {

			Matrix dataMatrixTEMP = subMatrix(i*5000, (i+1)*5000, 0, 4);
			if(i == 0) {
				dataMatrix = dataMatrixTEMP;
			} else {
				dataMatrix = dataMatrix.appendRight(dataMatrixTEMP);
			}
		}

		return dataMatrix;
	}

	public Matrix transpose() {  //m*n steps = O(mn)
		Matrix transposed = new Matrix(this.cols, this.rows);

		for (int r=0; r<rows; r++) { 		// m times
			for (int c=0; c<cols; c++) {	// n steps
				transposed.matrix[c][r]=matrix[r][c];
			}
		}

		return transposed;

	}


	public Matrix times(Matrix a) { //n*(a.n+mn) steps = n2*m+n*a.n steps 
		if (a.rows != this.cols) {
			throw new IllegalArgumentException("Matrix inner dimensions must agree.");
		}

		Matrix p = new Matrix(this.rows, a.cols);
		float[] Mcolj = new float[a.rows];

		for(int j = 0; j < a.cols; j++) {  //a.n times

			for(int k = 0; k < this.cols; k++) {  //n
				Mcolj[k] = a.matrix[k][j];
			}

			for(int i = 0; i < this.rows; i++) {  //m
				float[] rowi = this.matrix[i];
				float s = 0;
				for(int k = 0; k < this.cols; k++) {  //n
					s += rowi[k]*Mcolj[k];
				}
				p.matrix[i][j] = s;
			}
		}
		return p;
	}
	
	public double frobNorm() {
		double frobNorm = 0;
		
		for(int i = 0; i < this.rows; i++) {
			for(int j = 0; j < this.cols; j++) {		
				frobNorm += this.get(i, j)*this.get(i, j);
			}
		}
		
		return Math.sqrt(frobNorm);
	}
	
	public double average() {
		if(this.cols != 1) {
			throw new RuntimeException("Average can be calculated only on a vector!");
		}
			
		int sum = 0;
		for(int i = 0; i < this.rows; i++) {
			sum += this.matrix[i][0];
		}
		double average = sum/this.rows;
		return average;
	}
	
	public double covCoef(Matrix y) {
		if(this.cols != 1 || y.getCols() !=1 || this.rows != y.getRows()) {
			throw new RuntimeException("covCoef can be calculated only on a vector, and the two vector has to be the same size!");
		}
		double averageX = this.average();
		double averageY = y.average();
		
		double sum = 0;
		for(int i = 0; i < this.rows; i++) {
			double tempX = this.matrix[i][0]-averageX;
			double tempY = y.matrix[i][0]-averageY;
			sum+=tempX*tempY;
		}
		return sum;
	}
	
	public double devCoef() {
		if(this.cols != 1) {
			throw new RuntimeException("covCoef can be calculated only on a vector!");
		}
		double average = this.average();
		double sum = 0;
		for(int i = 0; i < this.rows; i++) {
			double temp = this.matrix[i][0]-average;
			sum+=temp*temp;
		}
		return sum;
	}
	
	public double pearsonCorrelationAbs(Matrix y) {
		
		double up = this.covCoef(y);
		double down = Math.sqrt(this.devCoef())*Math.sqrt(y.devCoef());
		
		double pearsonCorr = Math.abs(up/down);
		
		return  pearsonCorr;
	}
	

	public static void main(String[] args) {
//		Matrix m = new Matrix(5,2,1f);
		//			System.out.println("The original matrix: \n" + m+"\n\n");
		//			Matrix m1 = new Matrix(10,2,2f);
		//			m1.set(0, 0, 1);
		//			m1.set(0, 1, 2);
		//			m1.set(1, 0, 3);
		//			m1.set(1, 1, 4);
		//			m1.set(2, 0, 5);
		//			m1.set(2, 1, 6);
		//			System.out.println("The original matrix: \n" + m1+"\n\n");

//		Matrix m2 = m.appendReversed(m);


		Matrix x = new Matrix(10,1,1f);
		x.set(0, 0, 1);
		x.set(1, 0, 3);
		x.set(2, 0, 5);
		x.set(3, 0, 6);
		x.set(4, 0, 8);
		x.set(5, 0, 9);
		x.set(6, 0, 6);
		x.set(7, 0, 4);
		x.set(8, 0, 3);
		x.set(9, 0, 2);
			
		System.out.println("The original matrix: \n" + x+"\n\n");
		
		Matrix y = new Matrix(10,1,2f);
		
		y.set(0, 0, 2);
		y.set(1, 0, 5);
		y.set(2, 0, 6);
		y.set(3, 0, 6);
		y.set(4, 0, 7);
		y.set(5, 0, 7);
		y.set(6, 0, 5);
		y.set(7, 0, 3);
		y.set(8, 0, 1);
		y.set(9, 0, 1);

		System.out.println("The original matrix: \n" + y+"\n\n");
		
		double pearson = x.pearsonCorrelationAbs(y);
		
		System.out.println("A pearson corr: " + pearson);
		
		
		//			
		//			Matrix m3 = m2.subMatrix(0, 8, 0, 2);
		//			
		//			System.out.println("The original matrix: \n" + m3+"\n\n");
	}
}
